import re
import sys
import logging
from collections import namedtuple

from webob import Request, Response

from silk.default_config import CONFIG

logger = logging.getLogger("silk.apps")

Handler = namedtuple('Handler', ['regex', 'pattern', 'expanded_pattern', 'handler'])

class ImmediateResponse(Exception):
    def __init__(self, response):
        self.response = response

class WebApp:
    def __init__(self, handlers=None, middlewares=None, **config):
        self.middlewares = middlewares or {}
        for middleware_category in ('url', 'request', 'config', 'handler', 'response', 'exception'):
            self.middlewares.setdefault(middleware_category, ())
        handlers = handlers or {'^$': self.default_handler}
        self.handlers = []
        for k in handlers:
            if isinstance(k, Handler):
                self.handlers.append(k)
            else:
                original, expanded = k
                for um in self.middlewares['url']:
                    expanded = um(expanded)
                if original != expanded:
                    logger.info("Handler pattern expanded from %s to %s"%(original, expanded))
                self.handlers.append(Handler(regex=re.compile(expanded), original, expanded, handlers[k]))
        logger.debug("Handlers loaded:\n\n%s"%self.handlers)
        conf = CONFIG.copy()
        conf.update(config)
        self.config = conf

    def __call__(self, environ, start_response):
        try:
            try:
                request = Request(environ)
                logger.debug("Received request:\n\n%s"%request)
                request = self.run_middlewares('request', [request])
                config = self.config.copy()
                logger.debug("Running config middlewares:\t%s"%str(self.middlewares['config']))
                request, config = self.run_middlewares('config', [request, config])
                logger.debug("Beginning Request routing for path: %s"%request.path)
                for regex, pattern, expanded, handler in self.handlers:
                    match = regex.match(request.path)
                    if match:
                        logger.info("%s matches handler url of %s"%(request.path, pattern))
                        request, config, handler = self.run_middlewares('handler', [request,config,handler])
                        response = handler(request, config, **match.groupdict())
                        request, config, handler, response = self.run_middlewares('response', [request, config, handler, response])
                        start_response(response.status, response.headerlist)
                        return response.app_iter
                    raise ImmediateResponse(Response(status=404, body="<h1>404 Not Found</h1>"))
            except Exception, e:
                if isinstance(e, ImmediateResponse):
                    raise
                import traceback
                error_text = traceback.format_exc()
                logger.error("Encountered exception:\n\n%s"% error_text)
                self.run_middlewares('exception', [locals()])
                self.notify_admins()
                if not self.config.get('DEBUG', False):
                    error_text = '<h1>500 Internal Server Error</h1>'
                raise ImmediateResponse(Response(status=500, body=error_text))
        except ImmediateResponse, resp:
            start_response(resp.response.status, resp.response.headerlist)
            return resp.response.app_iter

    def run_middlewares(category, args):
        middlewares = self.middlewares[category]
        logger.debug("Running %s middlewares:\t%s"%(category, str(middlewares)))
        for middleware in self.middlewares[category]:
            ret = middleware(*args)
            if isinstance(ret, Response):
                logger.info("Middleware %s returned Response."%middleware)
                raise ImmediateResponse(ret)
            if ret is not None:
                args = ret
        return args

    def notify_admins(self):
        print "NOTIFY ADMINS"
    
    def serve(self, host='127.0.0.1', port=8080):
        from wsgiref.simple_server import make_server
        server = make_server(host, port, self)
        server.serve_forever()
